@@ -15,6 +15,7 @@
 
 import os
 import time
+from ZODB.utils import z64
 
 ##
 # ZEO storage server.
@@ -47,15 +48,6 @@ class StorageServer:
         zrpc.connection.Connection class.
         """
         self.rpc = rpc
-        
-        # Wait until we know what version the other side is using.
-        while rpc.peer_protocol_version is None:
-            time.sleep(0.1)
-
-        if rpc.peer_protocol_version == 'Z200':
-            self.lastTransaction = lambda: None
-            self.getInvalidations = lambda tid: None
-            self.getAuthProtocol = lambda: None
 
     def extensionMethod(self, name):
         return ExtensionMethodWrapper(self.rpc, name).call
@@ -95,7 +87,7 @@ class StorageServer:
 
     def lastTransaction(self):
         # Not in protocol version 2.0.0; see __init__()
-        return self.rpc.call('lastTransaction')
+        return self.rpc.call('lastTransaction') or z64
 
     ##
     # Return invalidations for all transactions after tid.
@@ -111,28 +103,26 @@ class StorageServer:
         return self.rpc.call('getInvalidations', tid)
 
     ##
-    # Check whether serial numbers s and sv are current for oid.
-    # If one or both of the serial numbers are not current, the
+    # Check whether a serial number is current for oid.
+    # If the serial number is not current, the
     # server will make an asynchronous invalidateVerify() call.
     # @param oid object id
-    # @param s serial number on non-version data
-    # @param sv serial number of version data or None
+    # @param s serial number
     # @defreturn async
 
-    def zeoVerify(self, oid, s, sv):
-        self.rpc.callAsync('zeoVerify', oid, s, sv)
+    def zeoVerify(self, oid, s):
+        self.rpc.callAsync('zeoVerify', oid, s)
 
     ##
-    # Check whether current serial number is valid for oid and version.
+    # Check whether current serial number is valid for oid.
     # If the serial number is not current, the server will make an
     # asynchronous invalidateVerify() call.
     # @param oid object id
-    # @param version name of version for oid
     # @param serial client's current serial number
     # @defreturn async
 
-    def verify(self, oid, version, serial):
-        self.rpc.callAsync('verify', oid, version, serial)
+    def verify(self, oid, serial):
+        self.rpc.callAsync('verify', oid, serial)
 
     ##
     # Signal to the server that cache verification is done.
@@ -166,34 +156,25 @@ class StorageServer:
             self.rpc.call('pack', t, wait)
 
     ##
-    # Return current data for oid.  Version data is returned if
-    # present.
+    # Return current data for oid.
     # @param oid object id
-    # @defreturn 5-tuple
-    # @return 5-tuple, current non-version data, serial number,
-    #         version name, version data, version data serial number
+    # @defreturn 2-tuple
+    # @return 2-tuple, current non-version data, serial number
     # @exception KeyError if oid is not found
 
     def zeoLoad(self, oid):
-        return self.rpc.call('zeoLoad', oid)
+        return self.rpc.call('zeoLoad', oid)[:2]
 
     ##
-    
-    # Return current data for oid in version, the tid of the
-    # transaction that wrote the most recent revision, and the name of
-    # the version for the data returned.  Note that if the object
-    # wasn't modified in the version, then the non-version data is
-    # returned and the returned version is an empty string.
+    # Return current data for oid, and the tid of the
+    # transaction that wrote the most recent revision.
     # @param oid object id
-    # @param version string, name of version
-    # @defreturn 3-tuple
-    # @return data, transaction id, version
-    #         where version is the name of the version the data came
-    #         from or "" for non-version data
+    # @defreturn 2-tuple
+    # @return data, transaction id
     # @exception KeyError if oid is not found
 
-    def loadEx(self, oid, version):
-        return self.rpc.call("loadEx", oid, version)
+    def loadEx(self, oid):
+        return self.rpc.call("loadEx", oid)
 
     ##
     # Return non-current data along with transaction ids that identify
@@ -213,14 +194,19 @@ class StorageServer:
     # @param oid object id
     # @param serial serial number that this transaction read
     # @param data new data record for oid
-    # @param version name of version or ""
     # @param id id of current transaction
     # @defreturn async
 
-    def storea(self, oid, serial, data, version, id):
-        self.rpc.callAsync('storea', oid, serial, data, version, id)
+    def storea(self, oid, serial, data, id):
+        self.rpc.callAsync('storea', oid, serial, data, id)
 
-    def storeBlob(self, oid, serial, data, blobfilename, version, txn):
+    def checkCurrentSerialInTransaction(self, oid, serial, id):
+        self.rpc.callAsync('checkCurrentSerialInTransaction', oid, serial, id)
+
+    def restorea(self, oid, serial, data, prev_txn, id):
+        self.rpc.callAsync('restorea', oid, serial, data, prev_txn, id)
+
+    def storeBlob(self, oid, serial, data, blobfilename, txn):
 
         # Store a blob to the server.  We don't want to real all of
         # the data into memory, so we use a message iterator.  This
@@ -235,13 +221,15 @@ class StorageServer:
                     break
                 yield ('storeBlobChunk', (chunk, ))
             f.close()
-            yield ('storeBlobEnd', (oid, serial, data, version, id(txn)))
+            yield ('storeBlobEnd', (oid, serial, data, id(txn)))
 
         self.rpc.callAsyncIterator(store())
 
-    def storeBlobShared(self, oid, serial, data, filename, version, id):
-        self.rpc.callAsync('storeBlobShared', oid, serial, data, filename, 
-                           version, id)
+    def storeBlobShared(self, oid, serial, data, filename, id):
+        self.rpc.callAsync('storeBlobShared', oid, serial, data, filename, id)
+
+    def deleteObject(self, oid, serial, id):
+        self.rpc.callAsync('deleteObject', oid, serial, id)
 
     ##
     # Start two-phase commit for a transaction
@@ -256,7 +244,7 @@ class StorageServer:
     # @defreturn async
 
     def tpc_begin(self, id, user, descr, ext, tid, status):
-        return self.rpc.call('tpc_begin', id, user, descr, ext, tid, status)
+        self.rpc.callAsync('tpc_begin', id, user, descr, ext, tid, status)
 
     def vote(self, trans_id):
         return self.rpc.call('vote', trans_id)
@@ -267,24 +255,15 @@ class StorageServer:
     def tpc_abort(self, id):
         self.rpc.callAsync('tpc_abort', id)
 
-    def abortVersion(self, src, id):
-        return self.rpc.call('abortVersion', src, id)
-
-    def commitVersion(self, src, dest, id):
-        return self.rpc.call('commitVersion', src, dest, id)
-
-    def history(self, oid, version, length=None):
+    def history(self, oid, length=None):
         if length is None:
-            return self.rpc.call('history', oid, version)
+            return self.rpc.call('history', oid)
         else:
-            return self.rpc.call('history', oid, version, length)
+            return self.rpc.call('history', oid, length)
 
     def record_iternext(self, next):
         return self.rpc.call('record_iternext', next)
 
-    def load(self, oid, version):
-        return self.rpc.call('load', oid, version)
-
     def sendBlob(self, oid, serial):
         return self.rpc.call('sendBlob', oid, serial)
 
@@ -294,17 +273,11 @@ class StorageServer:
     def loadSerial(self, oid, serial):
         return self.rpc.call('loadSerial', oid, serial)
 
-    def modifiedInVersion(self, oid):
-        return self.rpc.call('modifiedInVersion', oid)
-
     def new_oid(self):
         return self.rpc.call('new_oid')
 
-    def store(self, oid, serial, data, version, trans):
-        return self.rpc.call('store', oid, serial, data, version, trans)
-
-    def undo(self, trans_id, trans):
-        return self.rpc.call('undo', trans_id, trans)
+    def undoa(self, trans_id, trans):
+        self.rpc.callAsync('undoa', trans_id, trans)
 
     def undoLog(self, first, last):
         return self.rpc.call('undoLog', first, last)
@@ -312,14 +285,112 @@ class StorageServer:
     def undoInfo(self, first, last, spec):
         return self.rpc.call('undoInfo', first, last, spec)
 
-    def versionEmpty(self, vers):
-        return self.rpc.call('versionEmpty', vers)
+    def iterator_start(self, start, stop):
+        return self.rpc.call('iterator_start', start, stop)
 
-    def versions(self, max=None):
-        if max is None:
-            return self.rpc.call('versions')
+    def iterator_next(self, iid):
+        return self.rpc.call('iterator_next', iid)
+
+    def iterator_record_start(self, txn_iid, tid):
+        return self.rpc.call('iterator_record_start', txn_iid, tid)
+
+    def iterator_record_next(self, iid):
+        return self.rpc.call('iterator_record_next', iid)
+
+    def iterator_gc(self, iids):
+        return self.rpc.callAsync('iterator_gc', iids)
+
+    def server_status(self):
+        return self.rpc.call("server_status")
+
+    def set_client_label(self, label):
+        return self.rpc.callAsync('set_client_label', label)
+
+class StorageServer308(StorageServer):
+
+    def __init__(self, rpc):
+        if rpc.peer_protocol_version == 'Z200':
+            self.lastTransaction = lambda: z64
+            self.getInvalidations = lambda tid: None
+            self.getAuthProtocol = lambda: None
+
+        StorageServer.__init__(self, rpc)
+
+    def history(self, oid, length=None):
+        if length is None:
+            return self.rpc.call('history', oid, '')
         else:
-            return self.rpc.call('versions', max)
+            return self.rpc.call('history', oid, '', length)
+
+    def getInvalidations(self, tid):
+        # Not in protocol version 2.0.0; see __init__()
+        result = self.rpc.call('getInvalidations', tid)
+        if result is not None:
+            result = result[0], [oid for (oid, version) in result[1]]
+        return result
+
+    def verify(self, oid, serial):
+        self.rpc.callAsync('verify', oid, '', serial)
+
+    def loadEx(self, oid):
+        return self.rpc.call("loadEx", oid, '')[:2]
+
+    def storea(self, oid, serial, data, id):
+        self.rpc.callAsync('storea', oid, serial, data, '', id)
+
+    def storeBlob(self, oid, serial, data, blobfilename, txn):
+
+        # Store a blob to the server.  We don't want to real all of
+        # the data into memory, so we use a message iterator.  This
+        # allows us to read the blob data as needed.
+
+        def store():
+            yield ('storeBlobStart', ())
+            f = open(blobfilename, 'rb')
+            while 1:
+                chunk = f.read(59000)
+                if not chunk:
+                    break
+                yield ('storeBlobChunk', (chunk, ))
+            f.close()
+            yield ('storeBlobEnd', (oid, serial, data, '', id(txn)))
+
+        self.rpc.callAsyncIterator(store())
+
+    def storeBlobShared(self, oid, serial, data, filename, id):
+        self.rpc.callAsync('storeBlobShared', oid, serial, data, filename,
+                           '', id)
+
+    def zeoVerify(self, oid, s):
+        self.rpc.callAsync('zeoVerify', oid, s, None)
+
+    def iterator_start(self, start, stop):
+        raise NotImplementedError
+
+    def iterator_next(self, iid):
+        raise NotImplementedError
+
+    def iterator_record_start(self, txn_iid, tid):
+        raise NotImplementedError
+
+    def iterator_record_next(self, iid):
+        raise NotImplementedError
+
+    def iterator_gc(self, iids):
+        raise NotImplementedError
+
+def stub(client, connection):
+    start = time.time()
+    # Wait until we know what version the other side is using.
+    while connection.peer_protocol_version is None:
+        if time.time()-start > 10:
+            raise ValueError("Timeout waiting for protocol handshake")
+        time.sleep(0.1)
+
+    if connection.peer_protocol_version < 'Z309':
+        return StorageServer308(connection)
+    return StorageServer(connection)
+
 
 class ExtensionMethodWrapper:
     def __init__(self, rpc, name):
