@@ -25,11 +25,13 @@ from ZODB.tests import StorageTestBase, BasicStorage,  \
 from ZODB.tests.MinPO import MinPO
 from ZODB.tests.StorageTestBase import zodb_unpickle
 from ZODB.utils import p64, u64
-import asyncore
+from zope.testing import renormalizing
+
 import doctest
 import logging
 import os
 import persistent
+import re
 import shutil
 import signal
 import stat
@@ -45,6 +47,7 @@ import ZEO.tests.ConnectionTests
 import ZEO.zrpc.connection
 import ZODB
 import ZODB.blob
+import ZODB.tests.hexstorage
 import ZODB.tests.testblob
 import ZODB.tests.util
 import ZODB.utils
@@ -52,13 +55,12 @@ import zope.testing.setupstack
 
 logger = logging.getLogger('ZEO.tests.testZEO')
 
-ZEO.zrpc.connection.start_client_thread()
-
 class DummyDB:
     def invalidate(self, *args):
         pass
     def invalidateCache(*unused):
         pass
+    transform_record_data = untransform_record_data = lambda self, v: v
 
 
 class CreativeGetState(persistent.Persistent):
@@ -94,7 +96,8 @@ class MiscZEOTests:
 
     def checkZEOInvalidation(self):
         addr = self._storage._addr
-        storage2 = ClientStorage(addr, wait=1, min_disconnect_poll=0.1)
+        storage2 = self._wrap_client(
+            ClientStorage(addr, wait=1, min_disconnect_poll=0.1))
         try:
             oid = self._storage.new_oid()
             ob = MinPO('first')
@@ -125,7 +128,7 @@ class MiscZEOTests:
         addr = self._storage._addr
         storage2 = ClientStorage(addr)
         self.assert_(storage2.is_connected())
-        self.assertEquals(None, storage2.lastTransaction())
+        self.assertEquals(ZODB.utils.z64, storage2.lastTransaction())
         storage2.close()
 
         self._dostore()
@@ -196,13 +199,16 @@ class GenericTests(
             self.blob_cache_dir = tempfile.mkdtemp(
                 'blob_cache',
                 dir=os.path.abspath(os.getcwd()))
-        self._storage = ClientStorage(
+        self._storage = self._wrap_client(ClientStorage(
             zport, '1', cache_size=20000000,
             min_disconnect_poll=0.5, wait=1,
             wait_timeout=60, blob_dir=self.blob_cache_dir,
-            shared_blob_dir=self.shared_blob_dir)
+            shared_blob_dir=self.shared_blob_dir))
         self._storage.registerDB(DummyDB())
 
+    def _wrap_client(self, client):
+        return client
+
     def tearDown(self):
         self._storage.close()
         for server in self._servers:
@@ -243,10 +249,33 @@ class GenericTests(
         key = '%s:%s' % (self._storage._storage, self._storage._server_addr)
         self.assertEqual(self._storage.sortKey(), key)
 
+    def _do_store_in_separate_thread(self, oid, revid, voted):
+
+        def do_store():
+            store = ZEO.ClientStorage.ClientStorage(self._storage._addr)
+            try:
+                t = transaction.get()
+                store.tpc_begin(t)
+                store.store(oid, revid, 'x', '', t)
+                store.tpc_vote(t)
+                store.tpc_finish(t)
+            except Exception, v:
+                import traceback
+                print 'E'*70
+                print v
+                traceback.print_exception(*sys.exc_info())
+            finally:
+                store.close()
+
+        thread = threading.Thread(name='T2', target=do_store)
+        thread.setDaemon(True)
+        thread.start()
+        thread.join(voted and .1 or 9)
+        return thread
+
 class FullGenericTests(
     GenericTests,
     Cache.TransUndoStorageWithCache,
-    CommitLockTests.CommitLockUndoTests,
     ConflictResolution.ConflictResolvingStorage,
     ConflictResolution.ConflictResolvingTransUndoStorage,
     PackableStorage.PackableUndoStorage,
@@ -318,6 +347,16 @@ class FileStorageTests(FullGenericTests):
         </filestorage>
         """
 
+    _expected_interfaces = (
+        ('ZODB.interfaces', 'IStorageRestoreable'),
+        ('ZODB.interfaces', 'IStorageIteration'),
+        ('ZODB.interfaces', 'IStorageUndoable'),
+        ('ZODB.interfaces', 'IStorageCurrentRecordIteration'),
+        ('ZODB.interfaces', 'IExternalGC'),
+        ('ZODB.interfaces', 'IStorage'),
+        ('zope.interface', 'Interface'),
+        )
+
     def checkInterfaceFromRemoteStorage(self):
         # ClientStorage itself doesn't implement IStorageIteration, but the
         # FileStorage on the other end does, and thus the ClientStorage
@@ -327,18 +366,48 @@ class FileStorageTests(FullGenericTests):
         self.failUnless(ZODB.interfaces.IStorageIteration.providedBy(
             self._storage))
         # This is communicated using ClientStorage's _info object:
-        self.assertEquals(
-            (('ZODB.interfaces', 'IStorageRestoreable'),
-             ('ZODB.interfaces', 'IStorageIteration'),
-             ('ZODB.interfaces', 'IStorageUndoable'),
-             ('ZODB.interfaces', 'IStorageCurrentRecordIteration'),
-             ('ZODB.interfaces', 'IExternalGC'),
-             ('ZODB.interfaces', 'IStorage'),
-             ('zope.interface', 'Interface'),
-             ),
+        self.assertEquals(self._expected_interfaces,
             self._storage._info['interfaces']
             )
 
+class FileStorageHexTests(FileStorageTests):
+    _expected_interfaces = (
+        ('ZODB.interfaces', 'IStorageRestoreable'),
+        ('ZODB.interfaces', 'IStorageIteration'),
+        ('ZODB.interfaces', 'IStorageUndoable'),
+        ('ZODB.interfaces', 'IStorageCurrentRecordIteration'),
+        ('ZODB.interfaces', 'IExternalGC'),
+        ('ZODB.interfaces', 'IStorage'),
+        ('ZODB.interfaces', 'IStorageWrapper'),
+        ('zope.interface', 'Interface'),
+        )
+
+    def getConfig(self):
+        return """\
+        %import ZODB.tests
+        <hexstorage>
+        <filestorage 1>
+        path Data.fs
+        </filestorage>
+        </hexstorage>
+        """
+
+class FileStorageClientHexTests(FileStorageHexTests):
+
+    def getConfig(self):
+        return """\
+        %import ZODB.tests
+        <serverhexstorage>
+        <filestorage 1>
+        path Data.fs
+        </filestorage>
+        </serverhexstorage>
+        """
+
+    def _wrap_client(self, client):
+        return ZODB.tests.hexstorage.HexStorage(client)
+
+
 
 class MappingStorageTests(GenericTests):
     """ZEO backed by a Mapping storage."""
@@ -389,14 +458,17 @@ class HeartbeatTests(ZEO.tests.ConnectionTests.CommonSetupTearDown):
 
     def setUp(self):
         # Crank down the select frequency
-        self.__old_client_timeout = ZEO.zrpc.connection.client_timeout
-        ZEO.zrpc.connection.client_timeout = 0.1
-        ZEO.zrpc.connection.client_trigger.pull_trigger()
+        self.__old_client_timeout = ZEO.zrpc.client.client_timeout
+        ZEO.zrpc.client.client_timeout = self.__client_timeout
         ZEO.tests.ConnectionTests.CommonSetupTearDown.setUp(self)
 
+    __client_timeouts = 0
+    def __client_timeout(self):
+        self.__client_timeouts += 1
+        return .1
+
     def tearDown(self):
-        ZEO.zrpc.connection.client_timeout = self.__old_client_timeout
-        ZEO.zrpc.connection.client_trigger.pull_trigger()
+        ZEO.zrpc.client.client_timeout = self.__old_client_timeout
         ZEO.tests.ConnectionTests.CommonSetupTearDown.tearDown(self)
 
     def getConfig(self, path, create, read_only):
@@ -405,37 +477,24 @@ class HeartbeatTests(ZEO.tests.ConnectionTests.CommonSetupTearDown):
     def checkHeartbeatWithServerClose(self):
         # This is a minimal test that mainly tests that the heartbeat
         # function does no harm.
-        client_timeout_count = ZEO.zrpc.connection.client_timeout_count
         self._storage = self.openClientStorage()
-        time.sleep(1) # allow some time for the select loop to fire a few times
-        self.assert_(ZEO.zrpc.connection.client_timeout_count
-                     > client_timeout_count)
+        client_timeouts = self.__client_timeouts
+        forker.wait_until('got a timeout',
+                          lambda : self.__client_timeouts > client_timeouts
+                          )
         self._dostore()
 
-        if hasattr(os, 'kill'):
+        if hasattr(os, 'kill') and hasattr(signal, 'SIGKILL'):
             # Kill server violently, in hopes of provoking problem
             os.kill(self._pids[0], signal.SIGKILL)
             self._servers[0] = None
         else:
             self.shutdownServer()
 
-        for i in range(91):
-            # wait for disconnection
-            if not self._storage.is_connected():
-                break
-            time.sleep(0.1)
-        else:
-            raise AssertionError("Didn't detect server shutdown in 5 seconds")
-
-    def checkHeartbeatWithClientClose(self):
-        # This is a minimal test that mainly tests that the heartbeat
-        # function does no harm.
-        client_timeout_count = ZEO.zrpc.connection.client_timeout_count
-        self._storage = self.openClientStorage()
+        forker.wait_until('disconnected',
+                          lambda : not self._storage.is_connected()
+                          )
         self._storage.close()
-        time.sleep(1) # allow some time for the select loop to fire a few times
-        self.assert_(ZEO.zrpc.connection.client_timeout_count
-                     > client_timeout_count)
 
 
 class ZRPCConnectionTests(ZEO.tests.ConnectionTests.CommonSetupTearDown):
@@ -451,26 +510,26 @@ class ZRPCConnectionTests(ZEO.tests.ConnectionTests.CommonSetupTearDown):
             def writable(self):
                 raise SystemError("I'm evil")
 
-        log = []
-        ZEO.zrpc.connection.client_logger.critical = (
-            lambda m, *a, **kw: log.append((m % a, kw))
-            )
+        import zope.testing.loggingsupport
+        handler = zope.testing.loggingsupport.InstalledHandler(
+            'ZEO.zrpc.client')
 
-        ZEO.zrpc.connection.client_map[None] = Evil()
+        self._storage._rpc_mgr.map[None] = Evil()
 
         try:
-            ZEO.zrpc.connection.client_trigger.pull_trigger()
+            self._storage._rpc_mgr.trigger.pull_trigger()
         except DisconnectedError:
             pass
 
-        time.sleep(.1)
-        self.failIf(self._storage.is_connected())
-        self.assertEqual(len(ZEO.zrpc.connection.client_map), 1)
-        del ZEO.zrpc.connection.client_logger.critical
-        self.assertEqual(log[0][0], 'The ZEO client loop failed.')
-        self.assert_('exc_info' in log[0][1])
-        self.assertEqual(log[1][0], "Couldn't close a dispatcher.")
-        self.assert_('exc_info' in log[1][1])
+        forker.wait_until(
+            'disconnected',
+            lambda : not self._storage.is_connected()
+            )
+
+        log = str(handler)
+        handler.uninstall()
+        self.assert_("ZEO client loop failed" in log)
+        self.assert_("Couldn't close a dispatcher." in log)
 
     def checkExceptionLogsAtError(self):
         # Test the exceptions are logged at error
@@ -484,19 +543,21 @@ class ZRPCConnectionTests(ZEO.tests.ConnectionTests.CommonSetupTearDown):
 
         # This is a deliberately bogus call to get an exception
         # logged
-        self._storage._connection.handle_request('foo',0,'history',(1,2,3,4))
+        self._storage._connection.handle_request(
+            'foo', 0, 'history', (1, 2, 3, 4))
         # test logging
-        if __debug__:
-            level,message,kw = log[1]
-        else:
-            level,message,kw = log[0]
-        self.assertEqual(level,logging.ERROR)
-        self.failUnless(message.endswith(
+
+        for level, message, kw in log:
+            if message.endswith(
                 ') history() raised exception: history() takes at'
                 ' most 3 arguments (5 given)'
-                ))
-        self.assertEqual(kw,{'exc_info':True})
-        
+                ):
+                self.assertEqual(level,logging.ERROR)
+                self.assertEqual(kw,{'exc_info':True})
+                break
+        else:
+            self.fail("error not in log")
+
         # cleanup
         del conn.logger.log
 
@@ -737,6 +798,10 @@ class BlobWritableCacheTests(FullGenericTests, CommonBlobTests):
     blob_cache_dir = 'blobs'
     shared_blob_dir = True
 
+class FauxConn:
+    addr = 'x'
+    peer_protocol_version = ZEO.zrpc.connection.Connection.current_protocol
+
 class StorageServerClientWrapper:
 
     def __init__(self):
@@ -753,8 +818,8 @@ class StorageServerWrapper:
     def __init__(self, server, storage_id):
         self.storage_id = storage_id
         self.server = ZEO.StorageServer.ZEOStorage(server, server.read_only)
+        self.server.notifyConnected(FauxConn())
         self.server.register(storage_id, False)
-        self.server._thunk = lambda : None
         self.server.client = StorageServerClientWrapper()
 
     def sortKey(self):
@@ -776,8 +841,8 @@ class StorageServerWrapper:
         self.server.tpc_begin(id(transaction), '', '', {}, None, ' ')
 
     def tpc_vote(self, transaction):
-        self.server._restart()
-        self.server.vote(id(transaction))
+        vote_result = self.server.vote(id(transaction))
+        assert vote_result is None
         result = self.server.client.serials[:]
         del self.server.client.serials[:]
         return result
@@ -785,8 +850,14 @@ class StorageServerWrapper:
     def store(self, oid, serial, data, version_ignored, transaction):
         self.server.storea(oid, serial, data, id(transaction))
 
+    def send_reply(self, *args):        # Masquerade as conn
+        pass
+
+    def tpc_abort(self, transaction):
+        self.server.tpc_abort(id(transaction))
+
     def tpc_finish(self, transaction, func = lambda: None):
-        self.server.tpc_finish(id(transaction))
+        self.server.tpc_finish(id(transaction)).set_sender(0, self)
 
 
 def multiple_storages_invalidation_queue_is_not_insane():
@@ -859,6 +930,7 @@ Now we'll open a storage server on the data, simulating a restart:
     >>> fs = FileStorage('t.fs')
     >>> sv = StorageServer(('', get_port()), dict(fs=fs))
     >>> s = ZEOStorage(sv, sv.read_only)
+    >>> s.notifyConnected(FauxConn())
     >>> s.register('fs', False)
 
 If we ask for the last transaction, we should get the last transaction
@@ -950,6 +1022,8 @@ def tpc_finish_error():
     ...         pass
     ...     def close(self):
     ...         print 'connection closed'
+    ...     trigger = property(lambda self: self)
+    ...     pull_trigger = lambda self, func, *args: func(*args)
 
     >>> class ConnectionManager:
     ...     def __init__(self, addr, client, tmin, tmax):
@@ -1035,7 +1109,6 @@ def tpc_finish_error():
     connection closed
 
     >>> cs.close()
-    connection closed
     """
 
 def client_has_newer_data_than_server():
@@ -1075,13 +1148,13 @@ def client_has_newer_data_than_server():
     ... # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
     ZEO.ClientStorage CRITICAL client
     Client has seen newer transactions than server!
-    ZEO.zrpc ERROR (...) CW: error in notifyConnected (('localhost', ...))
+    ZEO.zrpc ERROR (...) CW: error in notifyConnected (('127.0.0.1', ...))
     Traceback (most recent call last):
     ...
     ClientStorageError: client Client has seen newer transactions than server!
     ZEO.ClientStorage CRITICAL client
     Client has seen newer transactions than server!
-    ZEO.zrpc ERROR (...) CW: error in notifyConnected (('localhost', ...))
+    ZEO.zrpc ERROR (...) CW: error in notifyConnected (('127.0.0.1', ...))
     Traceback (most recent call last):
     ...
     ClientStorageError: client Client has seen newer transactions than server!
@@ -1197,9 +1270,12 @@ def open_convenience():
 
 def client_asyncore_thread_has_name():
     """
+    >>> addr, _ = start_server()
+    >>> db = ZEO.DB(addr)
     >>> len([t for t in threading.enumerate()
-    ...      if t.getName() == 'ZEO.zrpc.connection'])
+    ...      if ' zeo client networking thread' in t.getName()])
     1
+    >>> db.close()
     """
 
 def runzeo_without_configfile():
@@ -1219,7 +1295,7 @@ def runzeo_without_configfile():
     ------
     --T INFO ZEO.runzeo () opening storage '1' using FileStorage
     ------
-    --T INFO ZEO.StorageServer () StorageServer created RW with storages 1RWt
+    --T INFO ZEO.StorageServer StorageServer created RW with storages 1RWt
     ------
     --T INFO ZEO.zrpc () listening on ...
     ------
@@ -1227,6 +1303,113 @@ def runzeo_without_configfile():
     testing exit immediately
     """
 
+def close_client_storage_w_invalidations():
+    r"""
+Invalidations could cause errors when closing client storages,
+
+    >>> addr, _ = start_server()
+    >>> writing = threading.Event()
+    >>> def mad_write_thread():
+    ...     global writing
+    ...     conn = ZEO.connection(addr)
+    ...     writing.set()
+    ...     while writing.isSet():
+    ...         conn.root.x = 1
+    ...         transaction.commit()
+    ...     conn.close()
+
+    >>> thread = threading.Thread(target=mad_write_thread)
+    >>> thread.setDaemon(True)
+    >>> thread.start()
+    >>> _ = writing.wait()
+    >>> time.sleep(.01)
+    >>> for i in range(10):
+    ...     conn = ZEO.connection(addr)
+    ...     _ = conn._storage.load('\0'*8)
+    ...     conn.close()
+
+    >>> writing.clear()
+    >>> thread.join(1)
+    """
+
+def convenient_to_pass_port_to_client_and_ZEO_dot_client():
+    """Jim hates typing
+
+    >>> addr, _ = start_server()
+    >>> client = ZEO.client(addr[1])
+    >>> client.__name__ == "('127.0.0.1', %s)" % addr[1]
+    True
+
+    >>> client.close()
+    """
+
+def test_server_status():
+    """
+    You can get server status using the server_status method.
+
+    >>> addr, _ = start_server(zeo_conf=dict(transaction_timeout=1))
+    >>> db = ZEO.DB(addr)
+    >>> import pprint
+    >>> pprint.pprint(db.storage.server_status(), width=1)
+    {'aborts': 0,
+     'active_txns': 0,
+     'commits': 1,
+     'conflicts': 0,
+     'conflicts_resolved': 0,
+     'connections': 1,
+     'loads': 1,
+     'lock_time': None,
+     'start': 'Tue May  4 10:55:20 2010',
+     'stores': 1,
+     'timeout-thread-is-alive': True,
+     'verifying_clients': 0,
+     'waiting': 0}
+
+    >>> db.close()
+    """
+
+def client_labels():
+    """
+When looking at server logs, for servers with lots of clients coming
+from the same machine, it can be very difficult to correlate server
+log entries with actual clients.  It's possible, sort of, but tedious.
+
+You can make this easier by passing a label to the ClientStorage
+constructor.
+
+    >>> addr, _ = start_server()
+    >>> db = ZEO.DB(addr, client_label='test-label-1')
+    >>> db.close()
+    >>> @wait_until
+    ... def check_for_test_label_1():
+    ...    for line in open('server-%s.log' % addr[1]):
+    ...        if 'test-label-1' in line:
+    ...            print line.split()[1:4]
+    ...            return True
+    ['INFO', 'ZEO.StorageServer', '(test-label-1']
+
+You can specify the client label via a configuration file as well:
+
+    >>> import ZODB.config
+    >>> db = ZODB.config.databaseFromString('''
+    ... <zodb>
+    ...    <zeoclient>
+    ...       server :%s
+    ...       client-label test-label-2
+    ...    </zeoclient>
+    ... </zodb>
+    ... ''' % addr[1])
+    >>> db.close()
+    >>> @wait_until
+    ... def check_for_test_label_2():
+    ...    for line in open('server-%s.log' % addr[1]):
+    ...        if 'test-label-2' in line:
+    ...            print line.split()[1:4]
+    ...            return True
+    ['INFO', 'ZEO.StorageServer', '(test-label-2']
+
+    """
+
 def invalidate_client_cache_entry_on_server_commit_error():
     """
 
@@ -1299,6 +1482,109 @@ But, if we abort, we'll get up to date data and we'll see the changes.
 
     """
 
+
+script_template = """
+import sys
+sys.path[:] = %(path)r
+
+%(src)s
+
+"""
+def generate_script(name, src):
+    open(name, 'w').write(script_template % dict(
+        exe=sys.executable,
+        path=sys.path,
+        src=src,
+        ))
+
+def runzeo_logrotate_on_sigusr2():
+    """
+    >>> port = get_port()
+    >>> open('c', 'w').write('''
+    ... <zeo>
+    ...    address %s
+    ... </zeo>
+    ... <mappingstorage>
+    ... </mappingstorage>
+    ... <eventlog>
+    ...    <logfile>
+    ...       path l
+    ...    </logfile>
+    ... </eventlog>
+    ... ''' % port)
+    >>> generate_script('s', '''
+    ... import ZEO.runzeo
+    ... ZEO.runzeo.main()
+    ... ''')
+    >>> import subprocess, signal
+    >>> p = subprocess.Popen([sys.executable, 's', '-Cc'], close_fds=True)
+    >>> wait_until('started',
+    ...   lambda : os.path.exists('l') and ('listening on' in open('l').read())
+    ...   )
+
+    >>> oldlog = open('l').read()
+    >>> os.rename('l', 'o')
+    >>> os.kill(p.pid, signal.SIGUSR2)
+
+    >>> wait_until('new file', lambda : os.path.exists('l'))
+    >>> s = ClientStorage(port)
+    >>> s.close()
+    >>> wait_until('See logging', lambda : ('Log files ' in open('l').read()))
+    >>> open('o').read() == oldlog # No new data in old log
+    True
+
+    # Cleanup:
+
+    >>> os.kill(p.pid, signal.SIGKILL)
+    >>> _ = p.wait()
+    """
+
+if sys.platform.startswith('win'):
+    del runzeo_logrotate_on_sigusr2
+
+if sys.version_info >= (2, 6):
+    import multiprocessing
+
+    def work_with_multiprocessing_process(name, addr, q):
+        conn = ZEO.connection(addr)
+        q.put((name, conn.root.x))
+        conn.close()
+
+    class MultiprocessingTests(unittest.TestCase):
+
+        layer = ZODB.tests.util.MininalTestLayer('work_with_multiprocessing')
+
+        def test_work_with_multiprocessing(self):
+            "Client storage should work with multi-processing."
+
+            # Gaaa, zope.testing.runner.FakeInputContinueGenerator has no close
+            if not hasattr(sys.stdin, 'close'):
+                sys.stdin.close = lambda : None
+            if not hasattr(sys.stdin, 'fileno'):
+                sys.stdin.fileno = lambda : -1
+
+            self.globs = {}
+            forker.setUp(self)
+            addr, adminaddr = self.globs['start_server']()
+            conn = ZEO.connection(addr)
+            conn.root.x = 1
+            transaction.commit()
+            q = multiprocessing.Queue()
+            processes = [multiprocessing.Process(
+                target=work_with_multiprocessing_process,
+                args=(i, addr, q))
+                         for i in range(3)]
+            _ = [p.start() for p in processes]
+            self.assertEqual(sorted(q.get(timeout=300) for p in processes),
+                             [(0, 1), (1, 1), (2, 1)])
+
+            _ = [p.join(30) for p in processes]
+            conn.close()
+            zope.testing.setupstack.tearDown(self)
+else:
+    class MultiprocessingTests(unittest.TestCase):
+        pass
+
 def quick_close_doesnt_kill_server():
     r"""
 
@@ -1324,8 +1610,8 @@ def quick_close_doesnt_kill_server():
     True
 
     >>> db.close()
-
     """
+
 def sync_connect_doesnt_hang():
     r"""
     >>> import threading
@@ -1349,7 +1635,6 @@ def sync_connect_doesnt_hang():
     >>> cm.thread.isAlive()
     False
     >>> ZEO.zrpc.client.ConnectThread = ConnectThread
-
     """
 
 def lp143344_extension_methods_not_lost_on_server_restart():
@@ -1364,10 +1649,10 @@ Make sure we don't lose exension methods on server restart.
     42
 
     >>> stop_server(adminaddr)
-    >>> forker.wait_until('not connected',
+    >>> wait_until('not connected',
     ...            lambda : not conn.db().storage.is_connected())
     >>> _ = start_server(addr=addr)
-    >>> forker.wait_until('connected', conn.db().storage.is_connected)
+    >>> wait_until('connected', conn.db().storage.is_connected)
 
     >>> conn.root.x
     1
@@ -1377,9 +1662,23 @@ Make sure we don't lose exension methods on server restart.
     >>> conn.close()
     """
 
+def can_use_empty_string_for_local_host_on_client():
+    """We should be able to spell localhost with ''.
+
+    >>> (_, port), _ = start_server()
+    >>> conn = ZEO.connection(('', port))
+    >>> conn.root()
+    {}
+    >>> conn.root.x = 1
+    >>> transaction.commit()
+
+    >>> conn.close()
+    """
+
 slow_test_classes = [
     BlobAdaptedFileStorageTests, BlobWritableCacheTests,
-    DemoStorageTests, FileStorageTests, MappingStorageTests,
+    MappingStorageTests, DemoStorageTests,
+    FileStorageTests, FileStorageHexTests, FileStorageClientHexTests,
     ]
 
 quick_test_classes = [
@@ -1449,7 +1748,11 @@ def test_suite():
     zeo = unittest.TestSuite()
     zeo.addTest(unittest.makeSuite(ZODB.tests.util.AAAA_Test_Runner_Hack))
     zeo.addTest(doctest.DocTestSuite(
-        setUp=forker.setUp, tearDown=zope.testing.setupstack.tearDown))
+        setUp=forker.setUp, tearDown=zope.testing.setupstack.tearDown,
+        checker=renormalizing.RENormalizing([
+            (re.compile(r"'start': '[^\n]+'"), 'start'),
+            ]),
+        ))
     zeo.addTest(doctest.DocTestSuite(ZEO.tests.IterationTests,
         setUp=forker.setUp, tearDown=zope.testing.setupstack.tearDown))
     zeo.addTest(doctest.DocFileSuite('registerDB.test'))
@@ -1471,6 +1774,8 @@ def test_suite():
     zeo.layer = ZODB.tests.util.MininalTestLayer('testZeo-misc')
     suite.addTest(zeo)
 
+    suite.addTest(unittest.makeSuite(MultiprocessingTests))
+
     # Put the heavyweights in their own layers
     for klass in slow_test_classes:
         sub = unittest.makeSuite(klass, "check")
