@@ -13,7 +13,7 @@
 ##############################################################################
 """Database connection support
 
-$Id: Connection.py 116715 2010-09-21 20:06:38Z jim $"""
+$Id: Connection.py 117091 2010-09-30 15:38:10Z jim $"""
 
 import logging
 import sys
@@ -88,6 +88,7 @@ class Connection(ExportImport, object):
         self._debug_info = ()
 
         self._db = db
+        self.large_record_size = db.large_record_size
 
         # historical connection
         self.before = before
@@ -104,7 +105,7 @@ class Connection(ExportImport, object):
             self._mvcc_storage = False
 
         self._normal_storage = self._storage = storage
-        self.new_oid = storage.new_oid
+        self.new_oid = db.new_oid
         self._savepoint_storage = None
 
         # Do we need to join a txn manager?
@@ -133,6 +134,10 @@ class Connection(ExportImport, object):
         # of this list are either in _cache or in _added.
         self._registered_objects = []
 
+        # ids and serials of objects for which readCurrent was called
+        # in a transaction.
+        self._readCurrent = {}
+
         # Dict of oid->obj added explicitly through add(). Used as a
         # preliminary cache until commit time when objects are all moved
         # to the real _cache. The objects are moved to _creating at
@@ -214,7 +219,7 @@ class Connection(ExportImport, object):
                             " added to a Connection.", obj)
         elif obj._p_jar is None:
             assert obj._p_oid is None
-            oid = obj._p_oid = self._storage.new_oid()
+            oid = obj._p_oid = self.new_oid()
             obj._p_jar = self
             if self._added_during_commit is not None:
                 self._added_during_commit.append(obj)
@@ -241,34 +246,30 @@ class Connection(ExportImport, object):
         if obj is not None:
             return obj
 
-        # This appears to be an MVCC violation because we are loading
-        # the must recent data when perhaps we shouldnt. The key is
-        # that we are only creating a ghost!
-        # A disadvantage to this optimization is that _p_serial cannot be
-        # trusted until the object has been loaded, which affects both MVCC
-        # and historical connections.
         p, serial = self._storage.load(oid, '')
         obj = self._reader.getGhost(p)
 
         # Avoid infiniate loop if obj tries to load its state before
         # it is added to the cache and it's state refers to it.
+        # (This will typically be the case for non-ghostifyable objects,
+        # like persistent caches.)
         self._pre_cache[oid] = obj
-        obj._p_oid = oid
-        obj._p_jar = self
-        obj._p_changed = None
-        obj._p_serial = serial
+        self._cache.new_ghost(oid, obj)
         self._pre_cache.pop(oid)
-        self._cache[oid] = obj
         return obj
 
     def cacheMinimize(self):
-        """Deactivate all unmodified objects in the cache."""
-        self._cache.minimize()
+        """Deactivate all unmodified objects in the cache.
+        """
+        for connection in self.connections.itervalues():
+            connection._cache.minimize()
 
     # TODO: we should test what happens when cacheGC is called mid-transaction.
     def cacheGC(self):
-        """Reduce cache size to target size."""
-        self._cache.incrgc()
+        """Reduce cache size to target size.
+        """
+        for connection in self.connections.itervalues():
+            connection._cache.incrgc()
 
     __onCloseCallbacks = None
     def onCloseCallback(self, f):
@@ -434,6 +435,7 @@ class Connection(ExportImport, object):
         if self._savepoint_storage is not None:
             self._abort_savepoint()
 
+        self._invalidate_creating()
         self._tpc_cleanup()
 
     def _abort(self):
@@ -450,6 +452,7 @@ class Connection(ExportImport, object):
                 del obj._p_oid
                 if obj._p_changed:
                     obj._p_changed = False
+                self._db.save_oid(oid)
             else:
 
                 # Note: If we invalidate a non-ghostifiable object
@@ -561,6 +564,14 @@ class Connection(ExportImport, object):
         else:
             self._commit(transaction)
 
+        for oid, serial in self._readCurrent.iteritems():
+            try:
+                self._storage.checkCurrentSerialInTransaction(
+                    oid, serial, transaction)
+            except ConflictError:
+                self._cache.invalidate(oid)
+                raise
+
     def _commit(self, transaction):
         """Commit changes to an object"""
 
@@ -641,7 +652,10 @@ class Connection(ExportImport, object):
                     and not hasattr(obj, '_p_resolveConflict')):
                     raise ConflictError(object=obj)
                 self._modified.append(oid)
+
             p = writer.serialize(obj)  # This calls __getstate__ of obj
+            if len(p) >= self.large_record_size:
+                warnings.warn(large_object_message % (obj.__class__, len(p)))
 
             if isinstance(obj, Blob):
                 if not IBlobStorage.providedBy(self._storage):
@@ -650,8 +664,12 @@ class Connection(ExportImport, object):
                         repr(self._storage))
                 if obj.opened():
                     raise ValueError("Can't commit with opened blobs.")
-                s = self._storage.storeBlob(oid, serial, p,
-                                            obj._uncommitted(),
+                blobfilename = obj._uncommitted()
+                if blobfilename is None:
+                    assert serial is not None # See _uncommitted
+                    self._modified.pop() # not modified
+                    continue
+                s = self._storage.storeBlob(oid, serial, p, blobfilename,
                                             '', transaction)
                 # we invalidate the object here in order to ensure
                 # that that the next attribute access of its name
@@ -678,32 +696,16 @@ class Connection(ExportImport, object):
             self._cache.update_object_size_estimation(oid, len(p))
             obj._p_estimated_size = len(p)
 
-            self._handle_serial(s, oid)
+            self._handle_serial(oid, s)
 
-    def _handle_serial(self, store_return, oid=None, change=1):
-        """Handle the returns from store() and tpc_vote() calls."""
+    def _handle_serial(self, oid, serial, change=True):
 
-        # These calls can return different types depending on whether
-        # ZEO is used.  ZEO uses asynchronous returns that may be
-        # returned in batches by the ClientStorage.  ZEO1 can also
-        # return an exception object and expect that the Connection
-        # will raise the exception.
+        # if we write an object, we don't want to check if it was read
+        # while current.  This is a convenient choke point to do this.
+        self._readCurrent.pop(oid, None)
 
-        # When conflict resolution occurs, the object state held by
-        # the connection does not match what is written to the
-        # database.  Invalidate the object here to guarantee that
-        # the new state is read the next time the object is used.
-
-        if not store_return:
+        if not serial:
             return
-        if isinstance(store_return, str):
-            assert oid is not None
-            self._handle_one_serial(oid, store_return, change)
-        else:
-            for oid, serial in store_return:
-                self._handle_one_serial(oid, serial, change)
-
-    def _handle_one_serial(self, oid, serial, change):
         if not isinstance(serial, str):
             raise serial
         obj = self._cache.get(oid, None)
@@ -757,6 +759,7 @@ class Connection(ExportImport, object):
             creating = self._creating
 
         for oid in creating:
+            self._db.save_oid(oid)
             o = self._cache.get(oid)
             if o is not None:
                 del self._cache[oid]
@@ -773,8 +776,16 @@ class Connection(ExportImport, object):
             vote = self._storage.tpc_vote
         except AttributeError:
             return
-        s = vote(transaction)
-        self._handle_serial(s)
+        try:
+            s = vote(transaction)
+        except ReadConflictError, v:
+            if v.oid:
+                self._cache.invalidate(v.oid)
+            raise
+
+        if s:
+            for oid, serial in s:
+                self._handle_serial(oid, serial)
 
     def tpc_finish(self, transaction):
         """Indicate confirmation that the transaction is done."""
@@ -812,6 +823,7 @@ class Connection(ExportImport, object):
     # pending invalidations regardless.  Of course this should only be
     # called at transaction boundaries.
     def _storage_sync(self, *ignored):
+        self._readCurrent.clear()
         sync = getattr(self._storage, 'sync', 0)
         if sync:
             sync()
@@ -861,16 +873,9 @@ class Connection(ExportImport, object):
         # load.  We can only be sure about invalidations after the
         # load.
 
-        # If an object has been invalidated, there are several cases
-        # to consider:
-        # 1. Check _p_independent()
-        # 2. Try MVCC
-        # 3. Raise ConflictError.
-
-        # Does anything actually use _p_independent()?  It would simplify
-        # the code if we could drop support for it.
-        # (BTrees.Length does.)
-
+        # If an object has been invalidated, among the cases to consider:
+        # - Try MVCC
+        # - Raise ConflictError.
 
         if self.before is not None:
             # Load data that was current before the time we have.
@@ -888,9 +893,7 @@ class Connection(ExportImport, object):
             if self._invalidatedCache:
                 raise ReadConflictError()
 
-            if (obj._p_oid in self._invalidated and
-                    not myhasattr(obj, "_p_independent")):
-                # If the object has _p_independent(), we will handle it below.
+            if (obj._p_oid in self._invalidated):
                 self._load_before_or_conflict(obj)
                 return
 
@@ -904,13 +907,8 @@ class Connection(ExportImport, object):
                 self._inv_lock.release()
 
             if invalid:
-                if myhasattr(obj, "_p_independent"):
-                    # This call will raise a ReadConflictError if something
-                    # goes wrong
-                    self._handle_independent(obj)
-                else:
-                    self._load_before_or_conflict(obj)
-                    return
+                self._load_before_or_conflict(obj)
+                return
 
         self._reader.setGhostState(obj, p)
         obj._p_serial = serial
@@ -959,25 +957,6 @@ class Connection(ExportImport, object):
 
         return True
 
-    def _handle_independent(self, obj):
-        # Helper method for setstate() handles possibly independent objects
-        # Call _p_independent(), if it returns True, setstate() wins.
-        # Otherwise, raise a ConflictError.
-
-        if obj._p_independent():
-            self._inv_lock.acquire()
-            try:
-                try:
-                    self._invalidated.remove(obj._p_oid)
-                except KeyError:
-                    pass
-            finally:
-                self._inv_lock.release()
-        else:
-            self._conflicts[obj._p_oid] = 1
-            self._register(obj)
-            raise ReadConflictError(object=obj)
-
     def register(self, obj):
         """Register obj with the current transaction manager.
 
@@ -1012,6 +991,10 @@ class Connection(ExportImport, object):
         if obj is not None:
             self._registered_objects.append(obj)
 
+    def readCurrent(self, ob):
+        assert ob._p_jar is self
+        assert ob._p_oid is not None and ob._p_serial is not None
+        self._readCurrent[ob._p_oid] = ob._p_serial
 
     # persistent.interfaces.IPersistentDatamanager
     ##########################################################################
@@ -1085,10 +1068,12 @@ class Connection(ExportImport, object):
         if getattr(self, '_reader', None) is not None:
             self._reader._cache = cache
 
-    def _releaseStorage(self):
-        """Tell the storage to release resources it's using"""
-        if self._mvcc_storage:
-            self._storage.release()
+    def _release_resources(self):
+        for c in self.connections.itervalues():
+            if c._mvcc_storage:
+                c._storage.release()
+            c._storage = c._normal_storage = None
+            c._cache = PickleCache(self, 0, 0)
 
     ##########################################################################
     # Python protocol
@@ -1166,7 +1151,6 @@ class Connection(ExportImport, object):
         src = self._savepoint_storage
         self._storage = self._normal_storage
         self._savepoint_storage = None
-
         try:
             self._log.debug("Committing savepoints of size %s", src.getSize())
             oids = src.index.keys()
@@ -1196,7 +1180,7 @@ class Connection(ExportImport, object):
                     s = self._storage.store(oid, serial, data,
                                             '', transaction)
 
-                self._handle_serial(s, oid, change=False)
+                self._handle_serial(oid, s, change=False)
         finally:
             src.close()
 
@@ -1398,3 +1382,23 @@ class RootConvenience(object):
         if len(names) > 60:
             names = names[:57].rsplit(' ', 1)[0] + ' ...'
         return "<root: %s>" % names
+
+large_object_message = """The %s
+object you're saving is large. (%s bytes.)
+
+Perhaps you're storing media which should be stored in blobs.
+
+Perhaps you're using a non-scalable data structure, such as a
+PersistentMapping or PersistentList.
+
+Perhaps you're storing data in objects that aren't persistent at
+all. In cases like that, the data is stored in the record of the
+containing persistent object.
+
+In any case, storing records this big is probably a bad idea.
+
+If you insist and want to get rid of this warning, use the
+large_record_size option of the ZODB.DB constructor (or the
+large-record-size option in a configuration file) to specify a larger
+size.
+"""
