@@ -13,7 +13,9 @@
 ##############################################################################
 """Unit tests for the Connection class."""
 
-from zope.testing import doctest
+from __future__ import with_statement
+
+import doctest
 import unittest
 import warnings
 
@@ -380,6 +382,36 @@ class UserMethodTests(unittest.TestCase):
         -1
         """
 
+def test_transaction_retry_convenience():
+    """
+    Simple test to verify integration with the transaction retry
+    helper my verifying that we can raise ConflictError and have it
+    handled properly.
+
+    This is an adaptation of the convenience tests in transaction.
+
+    >>> db = ZODB.tests.util.DB()
+    >>> conn = db.open()
+    >>> dm = conn.root()
+
+    >>> ntry = 0
+    >>> with transaction.manager:
+    ...      dm['ntry'] = 0
+
+    >>> import ZODB.POSException
+    >>> for attempt in transaction.manager.attempts():
+    ...     with attempt as t:
+    ...         t.note('test')
+    ...         print dm['ntry'], ntry
+    ...         ntry += 1
+    ...         dm['ntry'] = ntry
+    ...         if ntry % 3:
+    ...             raise ZODB.POSException.ConflictError()
+    0 0
+    0 1
+    0 2
+    """
+
 class InvalidationTests(unittest.TestCase):
 
     # It's harder to write serious tests, because some of the critical
@@ -556,6 +588,243 @@ def connection_root_convenience():
     <root: rather_long_name rather_long_name2 rather_long_name4 ...>
     """
 
+class proper_ghost_initialization_with_empty__p_deactivate_class(Persistent):
+    def _p_deactivate(self):
+        pass
+
+def proper_ghost_initialization_with_empty__p_deactivate():
+    """
+See https://bugs.launchpad.net/zodb/+bug/185066
+
+    >>> db = ZODB.tests.util.DB()
+    >>> conn = db.open()
+    >>> C = proper_ghost_initialization_with_empty__p_deactivate_class
+    >>> conn.root.x = x = C()
+    >>> conn.root.x.y = 1
+    >>> transaction.commit()
+
+    >>> conn2 = db.open()
+    >>> conn2.root.x._p_changed
+
+    >>> conn2.root.x.y
+    1
+
+    """
+
+def readCurrent():
+    r"""
+The connection's readCurrent method is called to provide a higher
+level of consistency in cases where an object if read to compute an
+update to a separate object.  When this is used, the
+checkCurrentSerialInTransaction method on the storage is called in
+2-phase commit.
+
+To demonstrate this, we'll create a storage and give it a test
+implementation of checkCurrentSerialInTransaction.
+
+    >>> import ZODB.MappingStorage
+    >>> store = ZODB.MappingStorage.MappingStorage()
+
+    >>> from ZODB.POSException import ReadConflictError
+    >>> bad = set()
+    >>> def checkCurrentSerialInTransaction(oid, serial, trans):
+    ...     print 'checkCurrentSerialInTransaction', `oid`
+    ...     if not trans == transaction.get(): print 'oops'
+    ...     if oid in bad:
+    ...         raise ReadConflictError(oid=oid)
+
+    >>> store.checkCurrentSerialInTransaction = checkCurrentSerialInTransaction
+
+Now, we'll use the storage as usual.  checkCurrentSerialInTransaction
+won't normally be called:
+
+    >>> db = ZODB.DB(store)
+    >>> conn = db.open()
+    >>> conn.root.a = ZODB.tests.util.P('a')
+    >>> conn.root.b = ZODB.tests.util.P('b')
+    >>> transaction.commit()
+
+If we call readCurrent for an object and we modify another object,
+then checkCurrentSerialInTransaction will be called for the object
+readCurrent was called on.
+
+    >>> conn.readCurrent(conn.root.a)
+    >>> conn.root.b.x = 0
+    >>> transaction.commit()
+    checkCurrentSerialInTransaction '\x00\x00\x00\x00\x00\x00\x00\x01'
+
+It doesn't matter how often we call readCurrent,
+checkCurrentSerialInTransaction will be called only once:
+
+    >>> conn.readCurrent(conn.root.a)
+    >>> conn.readCurrent(conn.root.a)
+    >>> conn.readCurrent(conn.root.a)
+    >>> conn.readCurrent(conn.root.a)
+    >>> conn.root.b.x += 1
+    >>> transaction.commit()
+    checkCurrentSerialInTransaction '\x00\x00\x00\x00\x00\x00\x00\x01'
+
+checkCurrentSerialInTransaction won't be called if another object
+isn't modified:
+
+
+    >>> conn.readCurrent(conn.root.a)
+    >>> transaction.commit()
+
+Or if the object it was called on is modified:
+
+    >>> conn.readCurrent(conn.root.a)
+    >>> conn.root.a.x = 0
+    >>> conn.root.b.x += 1
+    >>> transaction.commit()
+
+If the storage raises a conflict error, it'll be propigated:
+
+    >>> _ = str(conn.root.a) # do read
+    >>> bad.add(conn.root.a._p_oid)
+    >>> conn.readCurrent(conn.root.a)
+    >>> conn.root.b.x += 1
+    >>> transaction.commit()
+    Traceback (most recent call last):
+    ...
+    ReadConflictError: database read conflict error (oid 0x01)
+
+    >>> transaction.abort()
+
+The conflict error will cause the affected object to be invalidated:
+
+    >>> conn.root.a._p_changed
+
+The storage may raise it later:
+
+    >>> def checkCurrentSerialInTransaction(oid, serial, trans):
+    ...     if not trans == transaction.get(): print 'oops'
+    ...     print 'checkCurrentSerialInTransaction', `oid`
+    ...     store.badness = ReadConflictError(oid=oid)
+
+    >>> def tpc_vote(t):
+    ...     if store.badness:
+    ...        badness = store.badness
+    ...        store.badness = None
+    ...        raise badness
+
+    >>> store.checkCurrentSerialInTransaction = checkCurrentSerialInTransaction
+    >>> store.badness = None
+    >>> store.tpc_vote = tpc_vote
+
+It will still be propigated:
+
+    >>> _ = str(conn.root.a) # do read
+    >>> conn.readCurrent(conn.root.a)
+    >>> conn.root.b.x = +1
+    >>> transaction.commit()
+    Traceback (most recent call last):
+    ...
+    ReadConflictError: database read conflict error (oid 0x01)
+
+    >>> transaction.abort()
+
+The conflict error will cause the affected object to be invalidated:
+
+    >>> conn.root.a._p_changed
+
+Read checks don't leak accross transactions:
+
+    >>> conn.readCurrent(conn.root.a)
+    >>> transaction.commit()
+    >>> conn.root.b.x = +1
+    >>> transaction.commit()
+
+Read checks to work accross savepoints.
+
+    >>> conn.readCurrent(conn.root.a)
+    >>> conn.root.b.x = +1
+    >>> _ = transaction.savepoint()
+    >>> transaction.commit()
+    Traceback (most recent call last):
+    ...
+    ReadConflictError: database read conflict error (oid 0x01)
+
+    >>> transaction.abort()
+
+    >>> conn.readCurrent(conn.root.a)
+    >>> _ = transaction.savepoint()
+    >>> conn.root.b.x = +1
+    >>> transaction.commit()
+    Traceback (most recent call last):
+    ...
+    ReadConflictError: database read conflict error (oid 0x01)
+
+    >>> transaction.abort()
+
+    """
+
+def cache_management_of_subconnections():
+    """Make that cache management works for subconnections.
+
+When we use multi-databases, we open a connection in one database and
+access connections to other databases through it.  This test verifies
+thatcache management is applied to all of the connections.
+
+Set up a multi-database:
+
+    >>> db1 = ZODB.DB(None)
+    >>> db2 = ZODB.DB(None, databases=db1.databases, database_name='2',
+    ...               cache_size=10)
+    >>> conn1 = db1.open()
+    >>> conn2 = conn1.get_connection('2')
+
+Populate it with some data, more than will fit in the cache:
+
+    >>> for i in range(100):
+    ...     conn2.root()[i] = conn2.root().__class__()
+
+Upon commit, the cache is reduced to the cache size:
+
+    >>> transaction.commit()
+    >>> conn2._cache.cache_non_ghost_count
+    10
+
+Fill it back up:
+
+    >>> for i in range(100):
+    ...     _ = str(conn2.root()[i])
+    >>> conn2._cache.cache_non_ghost_count
+    101
+
+Doing cache GC on the primary also does it on the secondary:
+
+    >>> conn1.cacheGC()
+    >>> conn2._cache.cache_non_ghost_count
+    10
+
+Ditto for cache minimize:
+
+    >>> conn1.cacheMinimize()
+    >>> conn2._cache.cache_non_ghost_count
+    0
+
+
+Fill it back up:
+
+    >>> for i in range(100):
+    ...     _ = str(conn2.root()[i])
+    >>> conn2._cache.cache_non_ghost_count
+    101
+
+GC is done on reopen:
+
+    >>> conn1.close()
+    >>> db1.open() is conn1
+    True
+    >>> conn2 is conn1.get_connection('2')
+    True
+
+    >>> conn2._cache.cache_non_ghost_count
+    10
+
+    """
+
 class C_invalidations_of_new_objects_work_after_savepoint(Persistent):
     def __init__(self):
         self.settings = 1
@@ -573,7 +842,7 @@ def abort_of_savepoint_creating_new_objects_w_exotic_invalidate_doesnt_break():
     Really, when an object add is aborted, the object should be "removed" from
     the db and its invalidatuon method shouldm't even be called:
 
-    >>> conn = ZODB.connection('data.fs')
+    >>> conn = ZODB.connection(None)
     >>> conn.root.x = x = C_invalidations_of_new_objects_work_after_savepoint()
     >>> _ = transaction.savepoint()
     >>> x._p_oid
@@ -589,10 +858,6 @@ After the abort, the oid and jar are None:
     >>> x._p_oid
     >>> x._p_jar
 
-Cleanup:
-
-    >>> conn.close()
-
     """
 
 class Clp9460655(Persistent):
@@ -603,7 +868,7 @@ class Clp9460655(Persistent):
 
 def lp9460655():
     r"""
-    >>> conn = ZODB.connection('data.fs')
+    >>> conn = ZODB.connection(None)
     >>> root = conn.root()
     >>> Word = Clp9460655
 
@@ -625,11 +890,6 @@ def lp9460655():
     >>> print commonWords[1].id  # raises POSKeyError
     1
 
-    Cleanup:
-
-    >>> transaction.abort()
-    >>> conn.close()
-
     """
 
 def lp615758_transaction_abort_Incomplete_cleanup_for_new_objects():
@@ -640,7 +900,7 @@ reset "_p_changed" for new (i.e. "added") objects.
 
 >>> class P(Persistent): pass
 ...
->>> c = ZODB.connection('t.fs')
+>>> c = ZODB.connection(None)
 >>> obj = P()
 >>> c.add(obj)
 >>> obj.x = 1
@@ -653,6 +913,69 @@ False
 >>> c.close()
     """
 
+class Clp485456_setattr_in_getstate_doesnt_cause_multiple_stores(Persistent):
+
+    def __getstate__(self):
+        self.got = 1
+        return self.__dict__.copy()
+
+def lp485456_setattr_in_setstate_doesnt_cause_multiple_stores():
+    r"""
+    >>> C = Clp485456_setattr_in_getstate_doesnt_cause_multiple_stores
+    >>> conn = ZODB.connection(None)
+    >>> oldstore = conn._storage.store
+    >>> def store(oid, *args):
+    ...     print 'storing', repr(oid)
+    ...     return oldstore(oid, *args)
+    >>> conn._storage.store = store
+
+When we commit a change, we only get a single store call
+
+    >>> conn.root.x = C()
+    >>> transaction.commit()
+    storing '\x00\x00\x00\x00\x00\x00\x00\x00'
+    storing '\x00\x00\x00\x00\x00\x00\x00\x01'
+
+    >>> conn.add(C())
+    >>> transaction.commit()
+    storing '\x00\x00\x00\x00\x00\x00\x00\x02'
+
+We still see updates:
+
+    >>> conn.root.x.y = 1
+    >>> transaction.commit()
+    storing '\x00\x00\x00\x00\x00\x00\x00\x01'
+
+Not not non-updates:
+
+    >>> transaction.commit()
+
+Let's try some combinations with savepoints:
+
+    >>> conn.root.n = 0
+    >>> _ = transaction.savepoint()
+
+    >>> oldspstore = conn._storage.store
+    >>> def store(oid, *args):
+    ...     print 'savepoint storing', repr(oid)
+    ...     return oldspstore(oid, *args)
+    >>> conn._storage.store = store
+
+    >>> conn.root.y = C()
+    >>> _ = transaction.savepoint()
+    savepoint storing '\x00\x00\x00\x00\x00\x00\x00\x00'
+    savepoint storing '\x00\x00\x00\x00\x00\x00\x00\x03'
+
+    >>> conn.root.y.x = 1
+    >>> _  = transaction.savepoint()
+    savepoint storing '\x00\x00\x00\x00\x00\x00\x00\x03'
+
+    >>> transaction.commit()
+    storing '\x00\x00\x00\x00\x00\x00\x00\x00'
+    storing '\x00\x00\x00\x00\x00\x00\x00\x03'
+
+    >>> conn.close()
+    """
 
 class _PlayPersistent(Persistent):
     def setValueWithSize(self, size=0): self.value = size*' '
@@ -906,6 +1229,7 @@ class StubDatabase:
 
     def __init__(self):
         self.storage = StubStorage()
+        self.new_oid = self.storage.new_oid
 
     classFactory = None
     database_name = 'stubdatabase'
@@ -914,6 +1238,10 @@ class StubDatabase:
     def invalidate(self, transaction, dict_with_oid_keys, connection):
         pass
 
+    save_oid = lambda self, oid: None
+
+    large_record_size = 1<<30
+
 def test_suite():
     s = unittest.makeSuite(ConnectionDotAdd, 'check')
     s.addTest(doctest.DocTestSuite())
