@@ -18,6 +18,7 @@ storage to use for unit tests.  MappingStorage isn't sufficient.
 Since even a minimal storage has some complexity, we run standard
 storage tests against the test storage.
 """
+from __future__ import with_statement
 
 import bisect
 import threading
@@ -60,6 +61,8 @@ class MinimalMemoryStorage(BaseStorage, object):
         # _cur maps oid to current tid
         self._cur = {}
 
+        self._ltid = z64
+
     def isCurrent(self, oid, serial):
         return serial == self._cur[oid]
 
@@ -73,15 +76,13 @@ class MinimalMemoryStorage(BaseStorage, object):
     def _clear_temp(self):
         pass
 
-    def load(self, oid, version):
-        self._lock_acquire()
-        try:
+    def load(self, oid, version=''):
+        assert version == ''
+        with self._lock:
             assert not version
             tid = self._cur[oid]
             self.hook(oid, tid, '')
             return self._index[(oid, tid)], tid
-        finally:
-            self._lock_release()
 
     def _begin(self, tid, u, d, e):
         self._txn = Transaction(tid)
@@ -101,22 +102,15 @@ class MinimalMemoryStorage(BaseStorage, object):
         del self._txn
 
     def _finish(self, tid, u, d, e):
-        self._lock_acquire()
-        try:
+        with self._lock:
             self._index.update(self._txn.index)
             self._cur.update(self._txn.cur())
             self._ltid = self._tid
-        finally:
-            self._lock_release()
-
-    def lastTransaction(self):
-        return self._ltid
 
     def loadBefore(self, the_oid, the_tid):
         # It's okay if loadBefore() is really expensive, because this
         # storage is just used for testing.
-        self._lock_acquire()
-        try:
+        with self._lock:
             tids = [tid for oid, tid in self._index if oid == the_oid]
             if not tids:
                 raise KeyError(the_oid)
@@ -131,15 +125,9 @@ class MinimalMemoryStorage(BaseStorage, object):
             else:
                 end_tid = tids[j]
             return self._index[(the_oid, tid)], tid, end_tid
-        finally:
-            self._lock_release()
 
     def loadSerial(self, oid, serial):
-        self._lock_acquire()
-        try:
-            return self._index[(oid, serial)]
-        finally:
-            self._lock_release()
+        return self._index[(oid, serial)]
 
     def close(self):
         pass
@@ -154,6 +142,7 @@ class MinimalTestSuite(StorageTestBase.StorageTestBase,
                        ):
 
     def setUp(self):
+        StorageTestBase.StorageTestBase.setUp(self)
         self._storage = MinimalMemoryStorage()
 
     # we don't implement undo
